/*
 *  emovie.c
 *  This program animates output from IBIS immersed
 *  boundary simulations.  This program operates in
 *  double buffered mode for a flicker-free animation.
 *  The animation loops through the specified frames
 *  until the window is killed or the animation is stopped
 *  with the middle mouse button.
 *
 *  The program accepts input
 *
 *  From the mouse.
 *
 *    The left mouse button starts the animation.
 *    The right mouse button stops the animation.
 *
 *  From the key board.
 *
 *    q quits the animation.
 *    r reverses the animation.
 *    s starts the animation at the first frame.
 *
 *  Author:     David J. Eyre
 *              Department of Mathematics
 *              University of Utah
 *
 *  Date:       December 19, 1996
 *
 *  Copyright @ 1996, David J. Eyre
 *
 */
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/file.h>
#include "aux.h"
#include <fcntl.h>
#include <sys/time.h>

#define MAX 100
#define MAXC 1310720

long oldt = 0;
char base_name[MAX], color_map[MAX];
unsigned char image[MAXC], buf[MAX];

static GLint first_file = 0;
static GLint frame_rate = 100000;
static GLint height = 128;
static GLint ifile = 0;
static GLint increment = 1;
static GLint last_file = 30;
static GLint pixels = 32768;
static GLint skip_bytes = 4;
static GLint width = 256;
static GLint widthr = 256;
static GLint xleft;
static GLint yleft;
static GLint xright;
static GLint yright;
static GLfloat zoom = 1.0;

/*  
 *  Pause between frame display
 */

void wait_time(void){
    extern long oldt;
    struct timeval tp;
    long difft = 0;

    while (difft < frame_rate) {
      gettimeofday(&tp,(void *) NULL);
      difft = tp.tv_usec;
      difft = difft-oldt;
      if (difft <= 0) difft = difft+1000000;
    }
    gettimeofday(&tp,(void *) NULL);
    oldt = tp.tv_usec;
}

/*  
 *  display the images from disk files  - 
 *    wait a specified length of time first
 */
 
void display(void){

    extern char base_name[MAX];
    extern unsigned char image[MAXC], buf[MAX];

    unsigned char mimage[MAXC];
    int fp, err, i, j, k, q;
    char filename[MAX];

    sprintf(filename,"%s.%04d",base_name,ifile);
    if ((fp=open(filename,O_RDONLY,0)) != -1){
      if ( skip_bytes > 0 ) read(fp, buf, skip_bytes);
      if ( (err=read(fp,image,pixels)) == pixels ){
        q = 0;
        for ( i=yleft; i<yright; i++) {
          k = (i-1)*widthr;
          for(j=xleft; j<xright; j++) mimage[q++] = image[k+j];
	}
        glPixelZoom(zoom,zoom);
        glPixelStorei(GL_UNPACK_ALIGNMENT,1);
        wait_time();
        glDrawPixels(width,height,GL_COLOR_INDEX,GL_UNSIGNED_BYTE,mimage);
        glFlush();
      }
    }
    close(fp);
}

/*  
 *  Get the input data from the command line
 */

void getInData (int argc, char** argv)
{
    extern int optind;
    extern char *optarg;
    extern char base_name[MAX], color_map[MAX];

    int i, err=0;
    int uflag=0, vflag=0, xflag=0, yflag=0;
    int xl=0, yl=0, xr=0, yr=0;

    sprintf(base_name,"image");
    sprintf(color_map,"cmap");
/*  
 *  Parse the input arguments
 */
    if (argc == 1) err = 1;
    while ((i = getopt(argc,argv,"b:c:f:h:i:l:r:s:u:v:w:x:y:z:")) != EOF) {
      switch (i) {
        case 'b':
          sprintf(base_name,optarg);
          break;
        case 'c':
          sprintf(color_map,optarg);
          break;
        case 'f':
          first_file = atoi(optarg);
          break;
        case 'h':
          height = atoi(optarg);
          pixels = width*height;
          break;
        case 'i':
          increment = atoi(optarg);
          break;
        case 'l':
          last_file = atoi(optarg);
          break;
        case 'r':
          frame_rate = 1000000/atoi(optarg);
          break;
        case 's':
          skip_bytes = atoi(optarg);
          break;
        case 'u':
          uflag = 1;
          xl = atoi(optarg);
          break;
        case 'v':
          vflag = 1;
          yl = atoi(optarg);
          break;
        case 'w':
          width = atoi(optarg);
          widthr = width;
          pixels = width*height;
          break;
        case 'x':
          xflag = 1;
          xr = atoi(optarg);
          break;
        case 'y':
          yflag = 1;
          yr = atoi(optarg);
          break;
        case 'z':
          zoom = atof(optarg);
          break;
        default:
          err = 1;
      }
    }
   
    xleft  = 0;
    yleft  = 0;
    xright = width;
    yright = height;

    if (uflag==1) xleft = xl;
    if (vflag==1) yleft = yl;
    if (xflag==1) xright = xr;
    if (yflag==1) yright = yr;

    width = xright-xleft;
    height = yright-yleft;

/*  
 *  Oops, there was a boo-boo
 */
    if (err == 1) {
      printf("Usage: emovie [-h height] [-w width] ");
      printf("[-b base_name] [-f first_frame]\n");
      printf("   [-l last_frame] [-i increment] [-z zoom]\n");
      printf("   [-c color_map] [-r frame_rate] [-s skip_bytes] \n");
      printf("   [-u x_min] [-v y_min] [-x x_max] [-y y_max] \n");
      printf("Defaults: \n");
      printf("   -h 128 -w 256 -b image -f 0 -l 30");
      printf(" -i 1 -z 1.0 -c cmap -s 4 -r 10\n");
      printf("   -u 0 -v 0 -x 128 -y 256 \n\n");
      printf("While playing, the program accepts input.\n\n");
      printf("From the mouse.\n");
      printf("  The left mouse button starts the animation.\n");
      printf("  The right mouse button stops the animation.\n\n");
      printf("From the key board.\n");
      printf("  q quits the animation.\n");
      printf("  r reverses the animation.\n");
      printf("  s starts the animation at the first frame.\n");
      exit(0);
    }

}

 
/*  Initialize the color map and background color
 */

void myinit (void){

    FILE *fd;
    int i;
    float r,g,b;

    extern char color_map[MAX];

    fd = fopen(color_map,"r");
    for (i=0; i<=255; i++){
       fscanf(fd, "%e %e %e", &r, &g, &b);
       auxSetOneColor(i,r/255.,g/255.,b/255.);
    }
    fclose(fd);

    glClearIndex(0.0);

}

/*  Increment the movie file
 */
 
void nextFile (void){
    ifile = ifile+increment;
    if (ifile > last_file) ifile = first_file;
    if (ifile < first_file) ifile = last_file;
    display();
}

void startFunc (AUX_EVENTREC *event) { auxIdleFunc(nextFile); }

void stopFunc (AUX_EVENTREC *event) { auxIdleFunc(0); }

void exitFunc(void){ exit(0); }

void myReshape(int w, int h){ display(); }
 
void restartFunc (AUX_EVENTREC *event){
    ifile = first_file - 1;
    auxIdleFunc(nextFile);
}

void reverseFunc (AUX_EVENTREC *event){
    increment = -increment;
    auxIdleFunc(nextFile);
}
 
/*  Main Loop
 *  Get the input data from the command line.
 *  Open window with initial window size and title bar.
 *  Use double buffer and color index display mode. 
 *  Handle input events.  Loop indefinitely.
 */

int main(int argc, char** argv) {
    getInData (argc,argv);
    auxInitDisplayMode (AUX_SINGLE | AUX_INDEX);
    auxInitPosition (0, 0, width*zoom, height*zoom);
    auxInitWindow (argv[0]);
    myinit ();
    auxReshapeFunc (myReshape);
    auxIdleFunc (nextFile);
    auxMouseFunc (AUX_LEFTBUTTON, AUX_MOUSEDOWN, startFunc);
    auxMouseFunc (AUX_RIGHTBUTTON, AUX_MOUSEDOWN, stopFunc);
    auxKeyFunc (AUX_q, exitFunc);
    auxKeyFunc (AUX_s, restartFunc);
    auxKeyFunc (AUX_r, reverseFunc);
    auxMainLoop(display);
}
