/*
 *  ibisline.c
 *  This program displays a single output line graph
 *  from IBIS immersed boundary simulations.
 *
 *  See the IBIS manual for details on the input
 *  file format.
 *
 *  Author:     David J. Eyre
 *              Department of Mathematics
 *              University of Utah
 *
 *  Date:       December 15, 1996
 *
 *  Copyright @ 1996, David J. Eyre
 *
 */

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/file.h>
#include <fcntl.h>
#include "aux.h"
 
#define MAX 1024
#define screen_width 1280
#define screen_height 1024
#define MAXC screen_width*screen_height
#define min(a,b) (a < b) ? a : b
#define max(a,b) (a >= b) ? a : b

char color_map[MAX], out_file[MAX];
char data_file[MAX], title[MAX];
unsigned char image[MAXC], cimage[3*MAXC];
int r[256], b[256], g[256];

static GLint nfscalar;
static GLint nfvector;
static GLint nibpt;
static GLint nlink;
static GLint npmark;
static GLint npscalar;
static GLint npvector;

static GLint lwidth = 2;
static GLint pwidth = 4;
static GLint vwidth = 1;

static GLint nx;
static GLint ny;

static GLint ibisflag = 0;
static GLint ppmflag = 0;
static GLint psflag = 0;
static GLint tflag = 0;
static GLint xticflag = 0;
static GLint yticflag = 0;

static GLint size = 512;

static GLfloat aspect;
static GLfloat hgrid;
static GLfloat sun_time;
static GLfloat vmin;
static GLfloat vmax;
static GLfloat xmin;
static GLfloat xmax;
static GLfloat xlength;
static GLfloat ymin;
static GLfloat ymax;
static GLfloat ylength;


/* setup and graphically display data, return status */
void display(void){

  int cindex, i, j, k;
  float pi = 3.1415926535897, gamma = 0.2, hsgrid;
  float hgrid, theta, rad, u, v, x, y;
  float xal, xar, yal, yar;
  FILE *fd;

/* 
 * If no scalar field, then set background to index 0.
 */
  if ( nfscalar == 0 ) {
    if ( psflag == 0 ) {
      glIndexi(0);
      glBegin(GL_POLYGON);
        glVertex2f(xmin,ymin);
        glVertex2f(xmin,ymax);
        glVertex2f(xmax,ymax);
        glVertex2f(xmax,ymin);
      glEnd();
    }
  }

/* 
 * Box the domain in all postscript cases.
 */
  if ( psflag == 1 ) {
    sprintf(data_file,"%s.d.dat",out_file);
    remove(data_file);
    sprintf(data_file,"%s.v.dat",out_file);
    remove(data_file);
    sprintf(data_file,"%s.l.dat",out_file);
    remove(data_file);
    sprintf(data_file,"%s.m.dat",out_file);
    remove(data_file);
    sprintf(data_file,"%s.d.dat",out_file);
    fd = fopen(data_file,"w");
    fprintf(fd,"%e %e\n",xmin,ymin);
    fprintf(fd,"%e %e\n",xmin,ymax);
    fprintf(fd,"%e %e\n",xmax,ymax);
    fprintf(fd,"%e %e\n",xmax,ymin);
    fprintf(fd,"%e %e\n",xmin,ymin);
    fclose(fd);
  }
/* 
 * If a scalar field is present, plot it -- no action
 *  yet that plots scalar fields
 */

  if ( nfscalar == 0 ) goto SkipFieldScalar;
  fscanf (stdin, "%e", &hsgrid );
  for ( j=1; j<=nfscalar; j++ ) {
    fscanf (stdin, "%e %e %i", &x, &y, &cindex);
    if ( psflag == 0 ) {
      glIndexi(cindex);
      glBegin(GL_POLYGON);
        glVertex2f(x,y);
        glVertex2f(x+hsgrid,y);
        glVertex2f(x+hsgrid,y+hsgrid);
        glVertex2f(x,y+hsgrid);
      glEnd();
    }
  }
  SkipFieldScalar:

/* 
 * Outline the domain with color index 1.
 */
  if ( psflag == 0 ) {
    glIndexi(1);
    glLineWidth(lwidth+2);
    glBegin(GL_LINE_LOOP);
      glVertex2f(xmin,ymin);
      glVertex2f(xmin,ymax);
      glVertex2f(xmax,ymax);
      glVertex2f(xmax,ymin);
    glEnd();
  }

/* 
 * If a vector field is present, plot it
 */
  if ( nfvector == 0 ) goto SkipFieldVector;
  if ( psflag == 0 ) glLineWidth(vwidth);
  else {
    sprintf(data_file,"%s.v.dat",out_file);
    fd = fopen(data_file,"w");
  }
  fscanf (stdin, "%e", &vmin );
  for ( j=1; j<=nfvector; j++ ) {
    fscanf (stdin, "%e %e %e %e %i", 
                    &x, &y, &u, &v, &cindex);
    rad = sqrt((u*u)+(v*v));
    if ( rad >= vmin ) {
      if ( v >= 0 ) theta = acos(u/rad);
      if ( v < 0 ) theta = -acos(u/rad);
      rad = 5.*rad/8.;
      xal = x+(rad*cos(theta+gamma));
      yal = y+(rad*sin(theta+gamma));
      xar = x+(rad*cos(theta-gamma));
      yar = y+(rad*sin(theta-gamma));
      if ( psflag == 0 ) {
        glIndexi(cindex);
        glBegin(GL_LINE_STRIP);
          glVertex2f(x,y);
          glVertex2f(x+u,y+v);
        glEnd();
        glBegin(GL_LINE_STRIP);
          glVertex2f(xal,yal);
          glVertex2f(x+u,y+v);
          glVertex2f(xar,yar);
        glEnd();
      }
      else {
        fprintf(fd,"\n%e %e\n",x,y);
        fprintf(fd,"%e %e\n",x+u,y+v);
        fprintf(fd,"\n%e %e\n",xal,yal);
        fprintf(fd,"%e %e\n",x+u,y+v);
        fprintf(fd,"%e %e\n",xar,yar);
      }
    }
  }
  if ( psflag == 1 ) fclose(fd);
  SkipFieldVector:

/* 
 * If links are present, plot them
 */
  if ( psflag == 0 ) glLineWidth(lwidth);
  else {
    sprintf(data_file,"%s.l.dat",out_file);
    fd = fopen(data_file,"w");
  }
  for ( i=1; i<=nlink; i++ ) {
    fscanf (stdin, "%e %e %e %e %i", &x, &y, &u, &v, &cindex);
    if ( psflag == 0 ) {
      glIndexi(cindex);
      glBegin(GL_LINE_STRIP);
        glVertex2f(x,y);
        glVertex2f(u,v);
      glEnd();
    }
    else {
      fprintf(fd,"\n%e %e\n",x,y);
      fprintf(fd,"%e %e\n",u,v);
    }
  }    

/* 
 * If immersed boundary points are present, plot them
 */
  for ( i=1; i<=nibpt; i++ ) {
    if (i==1) fscanf (stdin, "%e", &rad);
    fscanf (stdin, "%e %e %i", &x, &y, &cindex);
    if ( psflag == 0 ) {
      glIndexi(cindex);
      glBegin(GL_LINE_LOOP);
        for ( k=1; k<=8; k++ ) {
          u = x+(rad*cos(k*pi/4.));
          v = y+(rad*sin(k*pi/4.));
          glVertex2f(u,v);
        }
      glEnd(); 
    }
    else {
      fprintf(fd,"\n%e %e\n",x+rad,y);
      for ( k=1; k<=8; k++ ) {
        u = x+(rad*cos(k*pi/4.));
        v = y+(rad*sin(k*pi/4.));
        fprintf(fd,"%e %e\n",u,v);
      }
    }
  }
  if ( psflag == 1 ) fclose(fd);

/* 
 * If immersed boundary point scalars are present, plot them
 */
  for ( i=1; i<=npscalar; i++ ) {
    fscanf (stdin, "%e %e %i", &x, &y, &cindex);
    if ( psflag == 0 ) {
      glIndexi(cindex);
      glBegin(GL_POLYGON);
        for ( k=1; k<=8; k++ ) {
          u = x+(rad*cos(k*pi/4.));
          v = y+(rad*sin(k*pi/4.));
          glVertex2f(u,v);
        }
      glEnd(); 
    }
  }


/* 
 * If point vectors field are present, plot them
 */
  if ( npvector == 0 ) goto SkipPointVector;
  if ( psflag == 0 ) glLineWidth(vwidth);
  else {
    sprintf(data_file,"%s.v.dat",out_file);
    fd = fopen(data_file,"a");
  }
  fscanf (stdin, "%e", &vmin);
  for ( i=1; i<=npvector; i++ ) {
    fscanf (stdin, "%e %e %e %e %i", &x, &y, &u, &v, &cindex);
    rad = sqrt((u*u)+(v*v));
    if ( rad >= vmin ) {
      if ( v >= 0 ) theta = acos(u/rad);
      if ( v < 0 ) theta = -acos(u/rad);
      rad = 5.*rad/8.;
      xal = x+(rad*cos(theta+gamma));
      yal = y+(rad*sin(theta+gamma));
      xar = x+(rad*cos(theta-gamma));
      yar = y+(rad*sin(theta-gamma));
      if ( psflag == 0 ) {
        glIndexi(cindex);
        glBegin(GL_LINE_STRIP);
          glVertex2f(x,y);
          glVertex2f(x+u,y+v);
        glEnd();
        glBegin(GL_LINE_STRIP);
          glVertex2f(xal,yal);
          glVertex2f(x+u,y+v);
          glVertex2f(xar,yar);
        glEnd();
      }
      else {
        fprintf(fd,"\n%e %e\n",x,y);
        fprintf(fd,"%e %e\n",x+u,y+v);
        fprintf(fd,"\n%e %e\n",xal,yal);
        fprintf(fd,"%e %e\n",x+u,y+v);
        fprintf(fd,"%e %e\n",xar,yar);
      }
    }
  }
  if ( psflag == 1 ) fclose(fd);
  SkipPointVector:

/* 
 * If fluid marker points are present, plot them
 */
  if ( npmark == 0 ) goto SkipFluidMarkers;
  if ( psflag == 0 ) glPointSize(pwidth);
  else {
    sprintf(data_file,"%s.m.dat",out_file);
    fd = fopen(data_file,"w");
  }

  for ( i=1; i<=npmark; i++ ) {
    if (i==1) fscanf (stdin, "%e", &rad);
    fscanf( stdin, "%e %e %i", &x, &y, &cindex ); 
    glIndexi(cindex);
    if ( psflag == 0 ) {
      glBegin(GL_POLYGON);
        for ( k=1; k<=8; k++ ) {
          u = x+(rad*cos(k*pi/4.));
          v = y+(rad*sin(k*pi/4.));
          glVertex2f(u,v);
        }
      glEnd(); 
    }
    else {
      fprintf(fd,"%e %e\n",x,y);
    }
  }
  if ( psflag == 1 ) fclose(fd);
  SkipFluidMarkers:

  if ( psflag == 0 ) {
    glFlush();
    glPixelStorei (GL_UNPACK_ALIGNMENT,1);
    glPixelStorei (GL_PACK_ALIGNMENT,1);
    glReadPixels (0,0,nx,ny,GL_COLOR_INDEX,GL_UNSIGNED_BYTE,image);
  }
  if ( psflag == 1 ) exit(0);
}

/*  
 *  Output a ppm format file.
 */
void OutputPPM(void){
  int i,j=0,k;

  for (i=1; i<=nx*ny; i++) {
    k = image[i];
    cimage[j++] = r[k];
    cimage[j++] = g[k];
    cimage[j++] = b[k];
    }
  fprintf(stdout, "P6\n");
  fprintf(stdout, "%i  %i\n", nx, ny);
  fprintf(stdout, "255\n");
  fwrite(cimage, nx*ny*3, 1, stdout);
  exit(0);
}

/*  
 *  Output an ibis color file.
 */
void OutputIBIS(void){
  fwrite(image, nx*ny, 1, stdout);
  exit(0);
}

void GetCmdLineData (int argc, char** argv) {

  extern int optind;
  extern char *optarg;

  FILE *fd;
  int c;
  char work[MAX],outext[MAX];
  int i, j=0, err=0, fflag=0, lim = MAX;
  sprintf(color_map,"ibisdat/cmap");

/*  
 *  Parse the input arguments
 */
  while ((i = getopt(argc,argv,"c:f:hl:o:p:s:v:")) != EOF) {
    switch (i) {
      case 'c':
        sprintf(color_map,optarg);
        break;
      case 'f':
        fflag = 1;
        sprintf(out_file,optarg);
        break;
      case 'h':
        err = 1;
        break;
      case 'l':
        lwidth = atoi(optarg);
        break;
      case 'o':
        sprintf(outext,optarg);
        sprintf(work,"ibis");
        if ( strncmp(outext,work,4) == 0 ) ibisflag = 1;
        sprintf(work,"ppm");
        if ( strncmp(outext,work,4) == 0 ) ppmflag = 1;
        sprintf(work,"pnm");
        if ( strncmp(outext,work,3) == 0 ) ppmflag = 1;
        sprintf(work,"ps");
        if ( strncmp(outext,work,2) == 0 ) psflag = 1;
        if ( ibisflag +psflag + ppmflag == 0 ) err = 1;
        break;
      case 'p':
        pwidth = atoi(optarg);
        break;
      case 's':
        size = atoi(optarg);
        break;
      case 'v':
        vwidth = atoi(optarg);
        break;
      default:
        err = 1;
    }
  }

  if ( psflag == 1 )
    if (fflag == 0) err = 1;
    if (fflag == 1) {
      fd = fopen(out_file,"r");
      fscanf ( fd, "%s\n", out_file );
      fgets( title, MAX, fd);
      while (--lim > 0 && title[j] != '\n' ) j++;
      title[j++] = '\0';
      fscanf ( fd, "%i", &xticflag );
      fscanf ( fd, "%i", &yticflag );
      sprintf(work,"null");
      if ( strncmp(title,work,4) == 0 ) tflag = 1;
      fclose(fd);
    }
/*  
 *  Error condition -- print usage
 */
  if (err == 1) {
    fprintf(stderr,"Synposis:  Displays IBIS line data\n");
    fprintf(stderr,"\n Usage: ibisline [-c color_map] [-f file_name]");
    fprintf(stderr,"  [-h] [-l lwidth] ");
    fprintf(stderr,"[-o output_format] [-p pwidth] [-s size]");
    fprintf(stderr,"  [-v vwidth]\n");
    fprintf(stderr,"    < infile [> outfile] \n");
    fprintf(stderr,"\n Required arguments:\n");
    fprintf(stderr,"    infile -- input file name\n");
    fprintf(stderr,"\n Optional arguments:\n");
    fprintf(stderr,"    -c color_map -- colormap index file");
    fprintf(stderr," (default = ibisdat/cmap)\n");
    fprintf(stderr,"    -f -- postscript options filename - required with");
    fprintf(stderr," .ps option\n");
    fprintf(stderr,"    -h -- this message\n");
    fprintf(stderr,"    -l lwidth -- line mark width for links ");
    fprintf(stderr,"(default = 1)\n");
    fprintf(stderr,"    -o output_format -- either ppm or ps\n");
    fprintf(stderr,"    -p pwidth -- fluid marker size\n");
    fprintf(stderr,"    -s size -- window size in pixels");
    fprintf(stderr," (default = 512)\n");
    fprintf(stderr,"    -v vwidth -- line mark width for vectors\n");
    fprintf(stderr,"    > output_file -- (default stdout)\n");
    fprintf(stderr,"\n Keyboard arguments:\n");
    fprintf(stderr,"    q - kill the window \n");
    fprintf(stderr,"    resizing not allowed \n");
    fprintf(stderr,"\n To create postscript type:\n");
    fprintf(stderr,"       gnuplot < file.gnu \n");
    exit(0);
  }
}

/* 
 * get the window dimensions, and data parameters
 */
void GetDomainData(void){

  fscanf( stdin, " %e %e %e %e %e", 
                   &xmin, &xmax, &ymin, &ymax, &sun_time);
  fscanf( stdin, " %i", &nfscalar);
  fscanf( stdin, " %i", &nfvector);
  fscanf( stdin, " %i", &nlink);
  fscanf( stdin, " %i", &nibpt);
  fscanf( stdin, " %i", &npscalar);
  fscanf( stdin, " %i", &npvector);
  fscanf( stdin, " %i", &npmark);

  if (xmax <= xmin){
    fprintf(stderr,"*******  ERROR   ********** \n");
    fprintf(stderr,"xmin must be less than xmax \n");
    exit(0);
  }
  if (ymax <= ymin){
    fprintf(stderr,"*******  ERROR   ********** \n");
    fprintf(stderr,"ymin must be less than ymax \n");
    exit(0);
  }

  xlength = xmax - xmin;
  ylength = ymax - ymin;
  aspect  = ylength/xlength;

  if ( xlength > ylength ) {
    nx = size;
    if ( nx > screen_width ) nx = screen_width;
    ny = nx*aspect;
    if ( ny > screen_height ){
      ny = screen_height;
      nx = ny/aspect;
    }
  }
  else {
    ny = size;
    if ( ny > screen_height ) ny = screen_height;
    nx = ny/aspect;
  }

  return;
}


/*
 *  Initialize the color map and background color
 */
void myinit (int argc, char** argv){

  FILE *fd;
  int i, err, fp;
  float red,green,blue;

/*  
 *  get the display going
 */
  auxInitDisplayMode ( AUX_SINGLE | AUX_INDEX );
  auxInitPosition (0, 0, nx, ny);
  auxInitWindow (argv[0]);

/*  
 *  get the colormap
 */
  fd = fopen(color_map,"r");
  for (i=0; i<=255; i++){
     fscanf(fd, "%e %e %e", &red, &green, &blue);
     r[i]=red; g[i]=green; b[i]=blue; 
     auxSetOneColor(i,red/255.,green/255.,blue/255.);
  }
  glClearIndex(0);
  glClear( GL_COLOR_BUFFER_BIT );
  glOrtho(xmin, xmax, ymin, ymax,-1.0, 1.0);
  display();
  if (ppmflag==1) OutputPPM();
  if (ibisflag==1) OutputIBIS();
  return;
}

/* write the gnuplot file */
void GnuFile(void){

  int k;
  char dash[10]; 
  FILE *fg;

  sprintf(dash,"'");
  sprintf(data_file,"%s.gnu",out_file);
  fg = fopen(data_file,"w");

  if (aspect >= 1 )
    fprintf(fg,"set terminal postscript portrait monochrome\n");
  if (aspect < 1 )
    fprintf(fg,"set terminal postscript landscape monochrome\n");

  fprintf(fg,"set output %s%s.ps%s\n",dash,out_file,dash);

  if ( tflag == 0 )
    fprintf(fg,"set title %s%s%s\n",dash,title,dash);

  fprintf(fg,"set multiplot\n");

  if ( xticflag == 0 ) fprintf(fg,"set noxtics\n");
  if ( yticflag == 0 ) fprintf(fg,"set noytics\n");

  fprintf(fg,"set size ratio %e\n",aspect);
  fprintf(fg,"set origin 0,0\n");

  fprintf(fg,"plot [x=%e:%e][%e:%e]", xmin,xmax,ymin,ymax);
  fprintf(fg," %s%s.d.dat%s notitle",dash,out_file,dash);
  fprintf(fg," with lines lw %i \n",lwidth+1);

  if ( (k=max(nlink,nibpt)) > 0) {
    fprintf(fg,"plot [x=%e:%e][%e:%e]", xmin,xmax,ymin,ymax);
    fprintf(fg," %s%s.l.dat%s notitle",dash,out_file,dash);
    fprintf(fg," with lines lw %i \n",lwidth);
  }

  if ( (k=max(nfvector,npvector)) > 0) {
    fprintf(fg,"plot [x=%e:%e][%e:%e]", xmin,xmax,ymin,ymax);
    fprintf(fg," %s%s.v.dat%s notitle",dash,out_file,dash);
    fprintf(fg," with lines lw %i \n",vwidth);
  }

  if ( npmark != 0 ) {
    fprintf(fg,"plot [x=%e:%e][%e:%e]", xmin,xmax,ymin,ymax);
    fprintf(fg," %s%s.m.dat%s notitle",dash,out_file,dash);
    fprintf(fg," with dots\n");
  }

  fprintf(fg,"set nomultiplot\n");
  fclose(fg);
  display();
}

void myReshape(int w, int h){ auxIdleFunc(0); }
void exFunc(void){ exit(0); }
void idle (void){ auxIdleFunc(0); }
void redisplay(void){
  glDrawPixels (nx,ny,GL_COLOR_INDEX,GL_UNSIGNED_BYTE,image);
}


/*  Main Program
 *  Get the input data from the command line.
 *  Open window with initial window size and title bar.
 *  Use double buffer and color index display mode. 
 *  Handle input events.  Loop indefinitely.
 */

int main(int argc, char** argv) {
    GetCmdLineData (argc,argv);
    GetDomainData ( );
    if ( psflag == 0 ) myinit (argc,argv);
    if ( psflag == 1 ) GnuFile ( );
    auxReshapeFunc (myReshape);
    auxIdleFunc (idle);
    auxKeyFunc (AUX_q, exFunc);
    auxKeyFunc (AUX_Q, exFunc);
    auxMainLoop (redisplay);
}
