/*
 *  ibextract.c
 *  This program extracts specified data from the
 *  from IBIS line data files generated in IBIS
 *  immersed boundary simulations.
 *
 *  See the IBIS manual for details on the input
 *  file format.
 *
 *  Author:     David J. Eyre
 *              Department of Mathematics
 *              University of Utah
 *
 *  Date:       August 21, 1997
 *
 *  Copyright @ 1997, David J. Eyre
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/file.h>
#include <fcntl.h>
 
#define MAX 1024
#define min(a,b) (a < b) ? a : b
#define max(a,b) (a >= b) ? a : b

FILE *fh;
int  file_ext;
char base_name[MAX], file_name[MAX];
char line[MAX];

float xmin, xmax, ymin, ymax;
int fsc, fvc, psc, pvc;
int fsd=1, fvd=1, psd=1, pvd=1;
int fscmin, fvcmin, pscmin, pvcmin;
int fscmax, fvcmax, pscmax, pvcmax;
float fsmin, fvmin, psmin, pvmin;
float fsmax, fvmax, psmax, pvmax;

int plink, pibpt, pmark;
int nx, ny;

int nibpt, nps, npv, nxyfs, nfs;
int nxyfv, nfv, nlinks, nmark;
float sun_time, hgrid;


/* getline: get line into s, return length */
int getline( char s[], int lim ){

  int c, i = 0;
      
  while (--lim > 0 && (c = getchar()) != EOF && c != '\n' )
    s[i++] = c;
  if ( c == '\n' )
    s[i++] = c;
  s[i] = '\0';

  return i;
}

/*  make sure input is provided  */
void CheckInputLine (int argc){
  if( argc == 2 ){
    fprintf(stderr," Usage: ibextract [-h] < cfile > outfile \n");
    fprintf(stderr,"    cfile -- control file for extracting data\n");
    fprintf(stderr,"    -h -- this message\n");
    fprintf(stderr,"    outfile -- output file for plotting\n");
    exit(0);
  }
  return;
}

/* get the window dimensions, and data parameters */
void GetInfileData(void){

  fscanf( stdin, " %s %i", base_name, &file_ext);
  fscanf( stdin, " %e %e %e %e", &xmin, &xmax, &ymin, &ymax);

  fscanf( stdin, " %i", &fsc );
  if ( fsc != 0 ) fscanf( stdin, " %e %e %i %i %i", 
              &fsmin, &fsmax, &fscmin, &fscmax, &fsd);

  fscanf( stdin, " %i", &fvc );
  if ( fvc != 0 ) fscanf( stdin, " %e %e %i %i %i", 
              &fvmin, &fvmax, &fvcmin, &fvcmax, &fvd);

  fscanf( stdin, " %i", &psc );
  if ( psc != 0 ) fscanf( stdin, " %e %e %i %i %i", 
              &psmin, &psmax, &pscmin, &pscmax, &psd);

  fscanf( stdin, " %i", &pvc );
  if ( pvc != 0 ) fscanf( stdin, " %e %e %i %i %i", 
              &pvmin, &pvmax, &pvcmin, &pvcmax, &pvd);

  fscanf( stdin, " %i", &plink);

  fscanf( stdin, " %i", &pibpt);

  fscanf( stdin, " %i", &pmark);

  return;
}


/* read the IBIS header file */
void ReadFileData(void){

  float x1,y1,x2,y2;

  sprintf (file_name,"%s.h.%4.4i",base_name,file_ext);
  fh = fopen(file_name,"r");
  fscanf( fh, " %e %e %e %e %e", &x1, &y1, &x2, &y2, &sun_time);
  fscanf( fh, " %i %i %e", &nx, &ny, &hgrid);
  fscanf( fh, " %i %i %i", &nibpt, &nps, &npv);
  fscanf( fh, " %i %i", &nxyfs, &nfs);
  fscanf( fh, " %i %i", &nxyfv, &nfv);
  fscanf( fh, " %i %i", &nlinks, &nmark);
  fclose(fh);

  return;

}

/* make sure the data provided is ok */
void CheckData (void){

  int err = 0, k;

  if ( fsmin > fsmax ) err = 1;
  if ( fvmin > fvmax ) err = 1;
  if ( psmin > psmax ) err = 1;
  if ( pvmin > pvmax ) err = 1;
  if ( err == 1 ){
    fprintf(stderr," \n*****  DATA ERROR *****\n");
    fprintf(stderr," Floating point min in the input file must be");
    fprintf(stderr," less than max\n");
    exit(0);
  }

  if ( fscmin > fscmax ) err = 1;
  if ( fvcmin > fvcmax ) err = 1;
  if ( pscmin > pscmax ) err = 1;
  if ( pvcmin > pvcmax ) err = 1;
  if ( err == 1 ){
    fprintf(stderr," \n*****  DATA ERROR *****\n");
    fprintf(stderr," Color min in the input file must be");
    fprintf(stderr," less than or equal to the max\n");
    exit(0);
  }

  if (fsd < 1) err = 1;
  if (fvd < 1) err = 1;
  if (psd < 1) err = 1;
  if (pvd < 1) err = 1;
  if ( err == 1 ){
    fprintf(stderr," \n*****  DATA ERROR *****\n");
    fprintf(stderr," Density variables must be integers greater");
    fprintf(stderr," than or equal to one\n");
    exit(0);
  }

  if ( fsc > nfs ) err = 1;
  if ( fvc > nfv ) err = 1;
  if ( psc > nps ) err = 1;
  if ( pvc > npv ) err = 1;
  if ( err == 1 ){
    fprintf(stderr," \n*****  DATA ERROR *****\n");
    fprintf(stderr," One of data columns requested does not exist\n");
    exit(0);
  }

  if ( fsc > 0 )
    if ( (k = nxyfs /(fsd*fsd)) == 0 ) err = 1;
  if ( fvc > 0 )
    if ( (k=nxyfv/(fvd*fvd)) == 0 ) err = 1;
  if ( psc > 0 )
    if ( (k=nps/psd) == 0 ) err = 1;
  if ( pvc > 0 )
    if ( (k=npv/pvd) == 0 ) err = 1;
  if ( err == 1 ){
    fprintf(stderr," \n*****  DATA ERROR *****\n");
    fprintf(stderr," One of the densities values is too large\n");
    exit(0);
  }

  return;
}

/* print the graph file header */
void PrintHeader(void){

  int k;

  fprintf( stdout, "%e %e %e %e %e\n", xmin, xmax, ymin, ymax, sun_time);

  if ( fsc == 0 ) fprintf( stdout, "0\n");
  k = (nx/fsd)*(ny/fsd);
  if ( fsc > 0 ) fprintf( stdout, "%i\n", k);
  
  if ( fvc == 0 ) fprintf( stdout, "0\n");
  k = (nx/fvd)*(ny/fvd);
  if ( fvc > 0 ) fprintf( stdout, "%i\n", k);

  if ( plink == 0 ) fprintf( stdout, "0\n");
  if ( plink > 0 ) fprintf( stdout, "%i\n", nlinks);

  if ( pibpt == 0 ) fprintf( stdout, "0\n");
  if ( pibpt > 0 ) fprintf( stdout, "%i\n", nibpt);

  if ( psc == 0 ) fprintf( stdout, "0\n");
  if ( psc > 0 ) fprintf( stdout, "%i\n", nibpt);

  if ( pvc == 0 ) fprintf( stdout, "0\n");
  if ( pvc > 0 ) fprintf( stdout, "%i\n", nibpt);

  if ( pmark == 0 ) fprintf( stdout, "0\n");
  if ( pmark > 0 ) fprintf( stdout, "%i\n", nmark);

  return;
}

/* print the graph file scalars */
void PrintFScalar(void){

  float x, y, s, sc;
  int i, j, k, id, jd, isc;

  sprintf (file_name,"%s.s.%4.4i",base_name,file_ext);
  fh = fopen(file_name,"r");

  for ( j=0; j<ny; j++ ) {
    jd = j%fsd;
    for ( i=0; i<nx; i++ ) {
      id = i%fsd;
      fscanf( fh, "%e %e", &x, &y );
      for ( k=1; k<fsc; k++ ) fscanf( fh, "%e", &s );
      fscanf( fh, "%e", &sc );
      for ( k=fsc+1; k<=nfs; k++ ) fscanf( fh, "%e", &s );
      k=max(id,jd);
      if ( k == 0 ){
        sc  = (sc-fsmin)/(fsmax-fsmin);
        isc = fscmin + ((fscmax-fscmin)*sc);
        fprintf( stdout, "%e %e %i\n", x, y, isc );
      }
    }
  }

  fclose(fh);
  return;
}

/* print the graph file field vectors */
void PrintFVectors(void){

  float u, uc, v, vc, normv, x, y;
  int i, j, k, ivc, id, jd;

  sprintf (file_name,"%s.v.%4.4i",base_name,file_ext);
  fh = fopen(file_name,"r");
  fprintf( stdout, "%e\n", fvmin*hgrid/fvmax );

  for ( j=0; j<ny; j++ ) {
    jd = j%fvd;
    for ( i=0; i<nx; i++ ) {
      id = i%fvd;
      fscanf( fh, "%e %e", &x, &y );
      for ( k=1; k<fvc; k++ ) fscanf( fh, "%e %e", &u, &v );
      fscanf( fh, "%e %e", &uc, &vc );
      for ( k=fvc+1; k<=nfv; k++ ) fscanf( fh, "%e %e", &u, &v );
      k=max(id,jd);
      if ( k == 0 ){
        normv = sqrt((uc*uc)+(vc*vc));
        normv = (normv-fvmin)/(fvmax-fvmin);
        ivc = fvcmin + ((fvcmax-fvcmin)*normv);
        u = uc*hgrid/fvmax; 
        v = vc*hgrid/fvmax;
        fprintf( stdout, "%e %e %e %e %i\n", x, y, u, v, ivc );
      }
    }
  }
  fclose(fh);
  return;
}

/* print the graph file links - someday do something with
 * the colors and s.  Fixed color index = 2 now.  */
void PrintLinks(void){

  float x, y, u, v;
  int j;
  
  sprintf (file_name,"%s.l.%4.4i",base_name,file_ext);
  fh = fopen(file_name,"r");
  for ( j=1; j<=nlinks; j++ ) {
    fscanf( fh, "%e %e %e %e", &x, &y, &u, &v );
    fprintf( stdout, "%e %e %e %e 3\n", x, y, u, v );
  }
  fclose(fh);

}

/* print the graph file immersed boundary points  */
void PrintIBPts(void){

  float x, y, u, v, s;
  int i, j;

/*
 *  print the immersed boundary points - someday, do something
 *  with the color of the points.  Fixed color index = 3 now.
 */  
  sprintf (file_name,"%s.i.%4.4i",base_name,file_ext);
  fh = fopen(file_name,"r");

  fprintf( stdout, "%e\n", hgrid/8);
  for ( j=1; j<=nibpt; j++ ) {
    fscanf( fh, "%e %e", &x, &y);
    for ( i=1; i<=npv; i++ ) fscanf( fh, "%e %e", &u, &v);
    for ( i=1; i<=nps; i++ ) fscanf( fh, "%e", &s);
    fprintf( stdout, "%e %e 4\n", x, y);
  }
  fclose(fh);
  return;
}

/* print the graph file point scalars  */
void PrintPScalar(void){

  float x, y, u, v, s, sc;
  int i, j, ic;

  sprintf (file_name,"%s.i.%4.4i",base_name,file_ext);
  fh = fopen(file_name,"r");
  for ( j=1; j<=nibpt; j++ ) {
    fscanf( fh, "%e %e", &x, &y);
    for ( i=1; i<=npv; i++ ) fscanf( fh, "%e %e", &u, &v);
    for ( i=1; i<psc; i++ ) fscanf( fh, "%e", &s);
    fscanf( fh, "%e", &sc);
    for ( i=psc+1; i<=nps; i++ ) fscanf( fh, "%e", &s);
    if ( (j-1)%psd == 0 ){
      sc = (sc-psmin)/(psmax-psmin);
      ic = pscmin + ((pscmax-pscmin)*sc);
      fprintf( stdout, "%e %e %i\n", x, y, ic );
    }
  }
  fclose(fh);
  return;
}

/* print the graph file point vector */
void PrintPVector(void){

  float x, y, u, v, s, uc, vc, h, normv;
  int i, j, ic, k;

  sprintf (file_name,"%s.i.%4.4i",base_name,file_ext);
  fh = fopen(file_name,"r");
  fprintf( stdout, "%e \n", pvmin*hgrid/pvmax );

  for ( j=0; j<nibpt; j++ ) {
    fscanf( fh, "%e %e", &x, &y);
    for ( i=1; i<pvc; i++ ) fscanf( fh, "%e %e", &u, &v);
    fscanf( fh, "%e %e", &uc, &vc);
    for ( i=pvc+1; i<=npv; i++ ) fscanf( fh, "%e %e", &u, &v);
    for ( i=1; i<=nps; i++ ) fscanf( fh, "%e", &s);
    if ( (k=(j-1)%pvd) == 0 ){
      normv = sqrt((uc*uc)+(vc*vc));
      normv = (normv-pvmin)/(pvmax-pvmin);
      ic = pvcmin + ((pvcmax-pvcmin)*normv);
      u = uc*hgrid/pvmax; 
      v = vc*hgrid/pvmax;
      fprintf( stdout, "%e %e %e %e %i\n", x, y, u, v, ic );
    }
  }
  fclose(fh);
  return;
}

/* print the graph file point vector */
void PrintMarkers(void){

  float x, y;
  int j, cindex;

  sprintf (file_name,"%s.m.%4.4i",base_name,file_ext);
  fh = fopen(file_name,"r");

  fprintf( stdout, "%e\n", hgrid/16);
  for ( j=0; j<nmark; j++ ) {
    fscanf( fh, "%e %e %i", &x, &y, &cindex);
    fprintf( stdout, "%e %e %i\n", x, y, cindex);
  }
  fclose(fh);
  return;
}

/*  Main Program  */
int main(int argc, char** argv) {
  CheckInputLine (argc);
  GetInfileData ( );
  ReadFileData ( );
  CheckData ( );
  PrintHeader ( );
  if ( fsc > 0 ) PrintFScalar( );
  if ( fvc > 0 ) PrintFVectors( );
  if ( plink > 0 ) PrintLinks( );
  if ( pibpt > 0 ) PrintIBPts( );
  if ( psc > 0 ) PrintPScalar( );
  if ( pvc > 0 ) PrintPVector( );
  if ( pmark > 0 ) PrintMarkers( );
  exit(0);
}
